
 /*
  * Unitex
  *
  * Copyright (C) 2001-2008 Universit Paris-Est Marne-la-Valle <unitex@univ-mlv.fr>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
  * version 2.1 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  * 
  * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
  *
  */

/*
 * Author: Burak Arslan (arslan@univ-mlv.fr, plq@gsulinux.org)
 *         This file contains the associative array (BDB-Judy thin abstraction layer) api implementation.
 */

/*
 * Uses a modified version of Judy-1.0.3 included in this package. It will *not*
 * work with other versions. This version was obtained via:
 *
 *     svn co http://svn.pugscode.org/pugs/third-party/judy/Judy-1.0.3 -r 10611
 */

/*
 * This library is tested with BDB 4.6.19. It may not work with older versions.
 */

/* 
 * Example uses of this library (iteration, insertion, deletion, etc) may be 
 * found in Collocation.cpp
 */

#include <stdlib.h>
#include <string.h>
#include "Array.h"

int array_init( Parray_t array ) {
#ifdef BDB

#define TMPPREFIX "/var/tmp/Colloc_"

	int ret;

    /* Initialize our handles */
	*array = NULL;
	DB_ENV *arrayE = NULL;
	DB_MPOOLFILE *arrayPF = NULL;
	
	int ret_t; 
	const char *db_name = "db";
	u_int32_t open_flags;

	/* Create the environment */
	ret = db_env_create(&arrayE, 0);
	if (ret != 0) {
		fprintf(stderr,"Error creating environment handle: %s\n", db_strerror(ret));
		exit(1);
	}
	
	open_flags =
		DB_CREATE     |  /* Create the environment if it does not exist */
		DB_INIT_MPOOL |  /* Initialize the memory pool (in-memory cache) */
		DB_PRIVATE;      /* Region files are not backed by the filesystem. 
		                  * Instead, they are backed by heap memory.  */

	/*
	 * Specify the size of the in-memory cache.
	 */
	ret = arrayE->set_cachesize(arrayE, 0, 200 * 1024 * 1024, 1); // according to bdb documentation, if 
	                                                                // this number is smaller than 500MB
	                                                                // a %25 overhead is added.
	if (ret != 0) {
		fprintf(stderr,"Error increasing the cache size: %s\n", db_strerror(ret));
		exit(1);
	}
	
	/*
	 * Now actually open the environment. Notice that the environment home
	 * directory is NULL. This is required for an in-memory only
	 * application.
	 */
	ret = arrayE->open(arrayE, NULL, open_flags, 0);
	if (ret != 0) {
		fprintf(stderr, "Error opening environment: %s\n", db_strerror(ret));
		exit(1);
	}
	
	/* Initialize the DB handle */
	ret = db_create(array, arrayE, 0);
	if (ret != 0) {
		arrayE->err(arrayE, ret, "Attempt to create db handle failed.");
		fprintf(stderr,"Bailing out...\n");
		exit(1);
	}
	
	/*
	 * Set the database open flags.
	 */
	char filename[32];
	sprintf(filename, "%s%d", TMPPREFIX, getpid() );
	open_flags = DB_CREATE | DB_TRUNCATE;
	ret = (*array)->open(*array, /* Pointer to the database */
				NULL,             /* Txn pointer */
				filename,         /* File name -- Must be NULL for inmemory! */
				db_name,          /* Logical db name */
				DB_BTREE,         /* Database type (using btree) */
				open_flags,       /* Open flags */
				0);               /* File mode. Using defaults */
	
	if (ret != 0) {
		arrayE->err(arrayE, ret, "Attempt to open db failed.");
		fprintf(stderr,"Bailing out...\n");
		exit(1);
	}

#else
	*array=NULL;

#endif

}

int array_get(Parray_t array, void *key, size_t keyL, void **data, size_t *dataL ) {
#ifdef BDB

	DBT arrayK, arrayD;
	int ret;

	memset (&arrayK, 0, sizeof(arrayK) );
	memset (&arrayD, 0, sizeof(arrayD) );

	arrayK.data  = key;
	arrayK.size  = keyL;

	ret = (*array)->get(*array, NULL, &arrayK, &arrayD, 0);
	if (ret) {
		*data=NULL;
		dataL=0;
        //fprintf(stderr, "%s\n", db_strerror(ret));
		return 1;
	}
	else {
		*data=arrayD.data;
		*dataL=arrayD.size;
		return 0;
	}
#else
	JUDYHSH(array);

	JHSG( arrayI, *array, key, keyL );

	*data=arrayI;
	if (arrayI) { 
		*dataL=sizeof(Word_t);
		return 0;
	}
	else {
		*dataL=0;
		return 1;
	}
#endif

}

int array_set(Parray_t array, void *key, size_t keyL, void *data, size_t dataL ) {

#ifdef BDB
	DBT arrayK, arrayD;
	int ret;

	arrayK.flags = 0;
	arrayK.data  = key;
	arrayK.size  = keyL;	

	arrayD.flags = 0;
	arrayD.data  = data;
	arrayD.size  = dataL;

	ret = (*array)->put(*array, NULL, &arrayK, &arrayD, 0);
	if (ret) {
        fprintf(stderr, "%s\n", db_strerror(ret));
		exit(1);
		return 1;
	}
	else {
		return 0;
	}
#else
	if (dataL > sizeof(Word_t)) {
		fprintf(stderr,"This array library only accepts word-size values, which is for this machine %d bytes.\n",sizeof(Pvoid_t));
		exit(1);
	}

	JUDYHSH(array);

	JHSI( arrayI, *array, key, keyL );
	*((Word_t*)arrayI)  = *((Word_t*)data);

	if (arrayI) { 
		return 0;
	}
	else {
		return 1;
	}
#endif

}

int array_free( Parray_t array, int contains_pointers ) {
	if ( contains_pointers ) {

	}
#ifdef BDB

	int ret, ret_t;
	DB_ENV *arrayE=NULL;

	/* Close our database handle, if it was opened. */
	if (array && *array) {
		arrayE = (*array)->get_env(*array);
		ret_t = (*array)->close(*array, 0);
		if (ret_t) {
			fprintf(stderr, "%s database close failed.\n", db_strerror(ret_t));
			ret = ret_t;
		}
		char filename[32];
		sprintf(filename, "%s%d", TMPPREFIX, getpid() );
	
		ret_t = arrayE->dbremove(arrayE,NULL,filename, NULL, 0);
		if (ret_t) {
			fprintf(stderr,"%s database remove failed.\n", db_strerror(ret_t));
			ret = ret_t;
		}
	}

	/* Close our environment, if it was opened. */
	if (arrayE) {
		ret_t = arrayE->close(arrayE, 0);
		if (ret_t) {
			fprintf(stderr,"environment close failed: %s\n", db_strerror(ret_t));
			ret = ret_t;
		}
	}

#else
	JUDYHSH(array);

	JHSFA(arrayAL,*array);

#endif

	*array=NULL;

}

int array_del( Parray_t array, void *key, size_t keyL ) {
#ifdef BDB
	DBT arrayK, arrayD;
	int ret;

	arrayK.flags = 0;
	arrayK.data  = key;
	arrayK.size  = keyL;

	arrayD.flags = 0;

	ret = (*array)->del(*array, NULL, &arrayK, 0);
	if (ret) {
		return 1;
	}
	else {
		return 0;
	}

#else
	JUDYHSH(array);

	JHSD( arrayR, *array, key, keyL );

	if (arrayR) { 
		return 0;
	}
	else {
		return 1;
	}

#endif

}

